"""
(*)~---------------------------------------------------------------------------
Pupil - eye tracking platform
Copyright (C) 2012-2020 Pupil Labs
Distributed under the terms of the GNU
Lesser General Public License (LGPL v3.0).
See LICENSE for license details.
---------------------------------------------------------------------------~(*)
"""
import abc
import collections
import logging
import typing as T
import uuid

import cv2
import numpy as np

from .coordinate_space import CoordinateSpace
from .corner import CornerId
from .marker import Marker, MarkerId, _Marker
from .orientation import SurfaceOrientation

logger = logging.getLogger(__name__)


SurfaceId = T.NewType("SurfaceId", str)


class Surface(abc.ABC):

    __versioned_subclasses = {}

    def __init_subclass__(cls):
        storage = Surface.__versioned_subclasses
        version = cls.version
        if version is None:
            raise ValueError(
                f'Surface subclass {cls.__name__} must overwrite class property "version"'
            )
        if version in storage:
            raise ValueError(
                f"Surface subclass {cls.__name__} defines an already registered version {version}"
            )
        storage[version] = cls
        return super().__init_subclass__()

    # ## Abstract members

    version = None  # type: ClassVar[int]

    @property
    @abc.abstractmethod
    def uid(self) -> SurfaceId:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def name(self) -> str:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def _registered_markers_by_uid_undistorted(self) -> T.Mapping[MarkerId, Marker]:
        raise NotImplementedError()

    @_registered_markers_by_uid_undistorted.setter
    @abc.abstractmethod
    def _registered_markers_by_uid_undistorted(
        self, value: T.Mapping[MarkerId, Marker]
    ):
        raise NotImplementedError()

    @property
    def registered_marker_uids(self) -> T.Set[MarkerId]:
        return set(self._registered_markers_by_uid_undistorted.keys())

    @property
    @abc.abstractmethod
    def orientation(self) -> SurfaceOrientation:
        raise NotImplementedError()

    @orientation.setter
    @abc.abstractmethod
    def orientation(self, value: SurfaceOrientation):
        raise NotImplementedError()

    # ## Update

    def _add_marker(self, marker_undistorted: Marker):
        if marker_undistorted.coordinate_space != CoordinateSpace.SURFACE_UNDISTORTED:
            raise ValueError(
                f"Expected marker_undistorted to be in undistorted surface space"
            )

        marker_uid = marker_undistorted.uid
        self._registered_markers_by_uid_undistorted[marker_uid] = marker_undistorted

    def _remove_marker(self, marker_uid: MarkerId):
        self._registered_markers_by_uid_undistorted[marker_uid] = None
        del self._registered_markers_by_uid_undistorted[marker_uid]

    def _move_corner(
        self,
        corner: CornerId,
        new_position_in_surface_space_undistorted: T.Tuple[float, float],
    ):
        self._registered_markers_by_uid_undistorted = self.__move_corner(
            corner=corner,
            registered_markers_by_uid=self._registered_markers_by_uid_undistorted,
            new_position_in_surface_space=new_position_in_surface_space_undistorted,
        )

    # ## Serialize

    @abc.abstractmethod
    def as_dict(self) -> dict:
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def from_dict(value: dict) -> "Surface":
        version = value["version"]
        surface_class = Surface.__versioned_subclasses.get(version, None)
        if surface_class is None:
            raise ValueError(
                f"No concrete implementation for Surface version {version}"
            )
        return surface_class.from_dict(value)

    # ## Factory

    @staticmethod
    def _create_surface_from_markers(
        name: str, markers: T.List[Marker]
    ) -> T.Optional["Surface"]:
        if not markers:
            return None

        if not all(
            m.coordinate_space == CoordinateSpace.IMAGE_UNDISTORTED for m in markers
        ):
            raise ValueError(f"TODO")

        markers = _check_markers_uniqueness(markers)
        marker_vertices_order = CornerId.all_corners()
        vertices = [m._vertices_in_order(order=marker_vertices_order) for m in markers]

        vertices_undistorted = np.array(vertices, dtype=np.float32)
        vertices_undistorted.shape = (-1, 2)

        marker_surface_coordinates_undistorted = _get_marker_surface_coordinates(
            vertices_undistorted
        )

        # Reshape to [marker, marker...]
        # where marker = [[u1, v1], [u2, v2], [u3, v3], [u4, v4]]
        marker_surface_coordinates_undistorted.shape = (-1, 4, 2)

        registered_markers_undistorted = {}

        # Add observations to library
        for marker, uv_undist in zip(markers, marker_surface_coordinates_undistorted):
            registered_markers_undistorted[marker.uid] = _Marker(
                uid=marker.uid,
                coordinate_space=CoordinateSpace.SURFACE_UNDISTORTED,
                vertices_by_corner_id=dict(
                    zip(marker_vertices_order, uv_undist.tolist())
                ),
            )

        # Create a new unique identifier for this surface
        uid = str(uuid.uuid4())
        uid = SurfaceId(uid)

        return _Surface_V2(
            uid=uid,
            name=name,
            registered_markers_undistorted=registered_markers_undistorted,
            orientation=SurfaceOrientation(),
        )

    # ## Private

    def __move_corner(
        self,
        corner: CornerId,
        registered_markers_by_uid: T.Mapping[MarkerId, Marker],
        new_position_in_surface_space: T.Tuple[float, float],
    ):
        order = CornerId.all_corners()

        old_corners = np.array([c.as_tuple() for c in order], dtype=np.float32)
        new_corners = np.array(
            [
                new_position_in_surface_space if c is corner else c.as_tuple()
                for c in order
            ],
            dtype=np.float32,
        )

        transform = cv2.getPerspectiveTransform(new_corners, old_corners)

        for marker_uid, marker in registered_markers_by_uid.items():
            old_vertices = np.asarray(
                marker._vertices_in_order(order=order), dtype=np.float32
            )
            new_vertices = cv2.perspectiveTransform(
                old_vertices.reshape((-1, 1, 2)), transform
            ).reshape((-1, 2))
            mapping = dict(zip(order, new_vertices.tolist()))
            registered_markers_by_uid[marker_uid] = _Marker(
                uid=marker_uid,
                coordinate_space=marker.coordinate_space,  # FIXME
                vertices_by_corner_id=mapping,
            )

        return registered_markers_by_uid


def _check_markers_uniqueness(markers: T.List[Marker]) -> T.List[Marker]:
    non_unique_markers_by_uid = collections.OrderedDict()

    for m in markers:
        non_unique = non_unique_markers_by_uid.get(m.uid, [])
        non_unique.append(m)
        non_unique_markers_by_uid[m.uid] = non_unique

    unique_markers = []

    for uid, non_unique in non_unique_markers_by_uid.items():
        if len(non_unique) > 1:
            logger.debug(f"Duplicate markers with uid {uid}.")
        unique_markers.append(non_unique[0])

    return unique_markers


def _get_marker_surface_coordinates(vertices: np.ndarray) -> np.ndarray:
    hull_distorted = _bounding_quadrangle(vertices)
    transform_candidate_image_to_surface = _get_transform_to_normalized_corners(
        hull_distorted
    )
    shape = vertices.shape
    vertices.shape = (-1, 1, 2)
    transform = cv2.perspectiveTransform(vertices, transform_candidate_image_to_surface)
    vertices.shape = shape
    return transform


def _get_transform_to_normalized_corners(vertices: np.ndarray) -> np.ndarray:
    corners = CornerId.all_corners(starting_with=CornerId.TOP_LEFT, clockwise=True)
    corners = [c.as_tuple() for c in corners]
    corners = np.array(corners, dtype=np.float32)
    return cv2.getPerspectiveTransform(vertices, corners)


def _bounding_quadrangle(vertices: np.ndarray):

    # According to OpenCV implementation, cv2.convexHull only accepts arrays with
    # 32bit floats (CV_32F) or 32bit signed ints (CV_32S).
    # See: https://github.com/opencv/opencv/blob/3.4/modules/imgproc/src/convhull.cpp#L137
    # See: https://github.com/pupil-labs/pupil/issues/1544
    vertices = np.asarray(vertices, dtype=np.float32)

    hull_points = cv2.convexHull(vertices, clockwise=False)

    # The convex hull of a list of markers must have at least 4 corners, since a
    # single marker already has 4 corners. If the convex hull has more than 4
    # corners we reduce that number with approximations of the hull.
    if len(hull_points) > 4:
        new_hull = cv2.approxPolyDP(hull_points, epsilon=1, closed=True)
        if new_hull.shape[0] >= 4:
            hull_points = new_hull

    if len(hull_points) > 4:
        curvature = abs(_GetAnglesPolyline(hull_points, closed=True))
        most_acute_4_threshold = sorted(curvature)[3]
        hull_points = hull_points[curvature <= most_acute_4_threshold]

    # Vertices space is flipped in y.  We need to change the order of the
    # hull_points vertices
    hull_points = hull_points[[1, 0, 3, 2], :, :]

    # Roll the hull_points vertices until we have the right orientation:
    # vertices space has its origin at the image center. Adding 1 to the
    # coordinates puts the origin at the top left.
    distance_to_top_left = np.sqrt(
        (hull_points[:, :, 0] + 1) ** 2 + (hull_points[:, :, 1] + 1) ** 2
    )
    bot_left_idx = np.argmin(distance_to_top_left) + 1
    hull_points = np.roll(hull_points, -bot_left_idx, axis=0)
    return hull_points


# From pupil_src/shared_modules/methods.py
def _GetAnglesPolyline(polyline, closed=False):
    """
    see: http://stackoverflow.com/questions/3486172/angle-between-3-points
    ported to numpy
    returns n-2 signed angles
    """

    points = polyline[:, 0]

    if closed:
        a = np.roll(points, 1, axis=0)
        b = points
        c = np.roll(points, -1, axis=0)
    else:
        a = points[0:-2]  # all "a" points
        b = points[1:-1]  # b
        c = points[2:]  # c points
    # ab =  b.x - a.x, b.y - a.y
    ab = b - a
    # cb =  b.x - c.x, b.y - c.y
    cb = b - c
    # float dot = (ab.x * cb.x + ab.y * cb.y); # dot product
    # print 'ab:',ab
    # print 'cb:',cb

    # float dot = (ab.x * cb.x + ab.y * cb.y) dot product
    # dot  = np.dot(ab,cb.T) # this is a full matrix mulitplication we only need the diagonal \
    # dot = dot.diagonal() #  because all we look for are the dotproducts of corresponding vectors (ab[n] and cb[n])
    dot = np.sum(
        ab * cb, axis=1
    )  # or just do the dot product of the correspoing vectors in the first place!

    # float cross = (ab.x * cb.y - ab.y * cb.x) cross product
    cros = np.cross(ab, cb)

    # float alpha = atan2(cross, dot);
    alpha = np.arctan2(cros, dot)
    return alpha * (180.0 / np.pi)  # degrees
    # return alpha #radians


# #### Concrete implementations


class _Surface_V2(Surface):

    version = 2

    @property
    def uid(self) -> SurfaceId:
        return self.__uid

    @property
    def name(self) -> str:
        return self.__name

    @property
    def _registered_markers_by_uid_undistorted(self) -> T.Mapping[MarkerId, Marker]:
        return self.__registered_markers_by_uid_undistorted

    @_registered_markers_by_uid_undistorted.setter
    def _registered_markers_by_uid_undistorted(
        self, value: T.Mapping[MarkerId, Marker]
    ):
        self.__registered_markers_by_uid_undistorted = value

    @property
    def orientation(self) -> SurfaceOrientation:
        return self.__orientation

    @orientation.setter
    def orientation(self, value: SurfaceOrientation):
        self.__orientation = value

    def as_dict(self) -> dict:
        registered_markers_undistorted = self._registered_markers_by_uid_undistorted
        registered_markers_undistorted = {
            k: v.as_dict() for k, v in registered_markers_undistorted.items()
        }
        return {
            "version": self.version,
            "uid": str(self.uid),
            "name": self.name,
            "registered_markers_undistorted": registered_markers_undistorted,
            "orientation": self.orientation.as_dict(),
        }

    @staticmethod
    def from_dict(value: dict) -> "Surface":
        try:
            actual_version = value["version"]
            expected_version = _Surface_V2.version
            assert (
                expected_version == actual_version
            ), f"Surface version missmatch; expected {expected_version}, but got {actual_version}"

            registered_markers_undistorted = value["registered_markers_undistorted"]
            registered_markers_undistorted = {
                k: Marker.from_dict(v)
                for k, v in registered_markers_undistorted.items()
            }

            orientation_dict = value.get("orientation", None)
            if orientation_dict:
                orientation = SurfaceOrientation.from_dict(orientation_dict)
            else:
                # use default if surface was saved as dict before this change
                orientation = SurfaceOrientation()

            return _Surface_V2(
                uid=SurfaceId(value["uid"]),
                name=value["name"],
                registered_markers_undistorted=registered_markers_undistorted,
                orientation=orientation,
            )
        except Exception as err:
            raise ValueError(err)

    def __init__(
        self,
        uid: SurfaceId,
        name: str,
        registered_markers_undistorted: T.Mapping[MarkerId, Marker],
        orientation: SurfaceOrientation,
    ):
        self.__uid = uid
        self.__name = name
        self.__registered_markers_by_uid_undistorted = registered_markers_undistorted
        self.__orientation = orientation
        assert all(
            m.coordinate_space == CoordinateSpace.SURFACE_UNDISTORTED
            for m in registered_markers_undistorted.values()
        )



"""
(*)~---------------------------------------------------------------------------
Pupil - eye tracking platform
Copyright (C) 2012-2020 Pupil Labs
Distributed under the terms of the GNU
Lesser General Public License (LGPL v3.0).
See LICENSE for license details.
---------------------------------------------------------------------------~(*)
"""
import enum
import typing as T

import cv2
import numpy as np

from .location import SurfaceLocation
from .surface import SurfaceId


class SurfaceHeatmap:
    class ColorFormat(enum.Enum):
        RGB = enum.auto()

        @property
        def channel_count(self) -> int:
            if self == SurfaceHeatmap.ColorFormat.RGB:
                return 3
            raise NotImplementedError()

    @staticmethod
    def _create_surface_heatmap(
        points_in_image_space: T.List[T.Tuple[int, int]], location: SurfaceLocation
    ):
        surface_heatmap = SurfaceHeatmap(surface_uid=location.surface_uid)
        surface_heatmap._add_points(
            points_in_image_space=points_in_image_space, location=location
        )
        return surface_heatmap

    def __init__(self, surface_uid: SurfaceId):
        self.__surface_uid = surface_uid
        self.__points_in_surface_space_numpy = []
        self._invalidate_cached_computations()

    def _add_points(
        self,
        points_in_image_space: T.List[T.Tuple[int, int]],
        location: SurfaceLocation,
    ):
        points_in_image_space_numpy = np.asarray(
            points_in_image_space, dtype=np.float32
        )
        new_points_in_surface_space_numpy = location._map_from_image_to_surface(
            points=np.asarray(points_in_image_space_numpy, dtype=np.float32)
        )

        for p in new_points_in_surface_space_numpy:
            self.__points_in_surface_space_numpy.append(p)

    def image(
        self,
        size: T.Tuple[int, int],
        color_format: T.Optional["SurfaceHeatmap.ColorFormat"] = None,
    ) -> np.ndarray:

        if color_format is None:
            color_format = SurfaceHeatmap.ColorFormat.RGB
        elif not isinstance(color_format, SurfaceHeatmap.ColorFormat):
            raise ValueError(
                f"color_format must be an instance of SurfaceHeatmap.ColorFormat"
            )

        cache_key = (size, color_format)
        heatmap_resolution = 31
        heatmap_blur_factor = 0.0

        if cache_key not in self.__heatmap_image_by_size_and_color_format:
            heatmap_data = np.asarray(self.__points_in_surface_space_numpy)

            aspect_ratio = size[1] / size[0]
            grid = (
                max(1, int(heatmap_resolution * aspect_ratio)),
                int(heatmap_resolution),
            )

            if len(heatmap_data) > 0:
                xvals = heatmap_data[:, 0]
                yvals = 1.0 - heatmap_data[:, 1]
                histogram, *edges = np.histogram2d(
                    yvals, xvals, bins=grid, range=[[0, 1.0], [0, 1.0]], normed=False
                )
                filter_h = 19 + heatmap_blur_factor * 15
                filter_w = filter_h * aspect_ratio
                filter_h = int(filter_h) // 2 * 2 + 1
                filter_w = int(filter_w) // 2 * 2 + 1

                histogram = cv2.GaussianBlur(histogram, (filter_h, filter_w), 0)
                histogram_max = histogram.max()
                histogram *= (255.0 / histogram_max) if histogram_max else 0.0
                histogram = histogram.astype(np.uint8)
            else:
                histogram = np.zeros(grid, dtype=np.uint8)

            histogram = cv2.applyColorMap(histogram, cv2.COLORMAP_JET)

            if color_format == SurfaceHeatmap.ColorFormat.RGB:
                heatmap = np.ones((*grid, color_format.channel_count), dtype=np.uint8)
                heatmap[:, :, 0] = histogram[:, :, 2]  # red
                heatmap[:, :, 1] = histogram[:, :, 1]  # green
                heatmap[:, :, 2] = histogram[:, :, 0]  # blue
            else:
                raise ValueError(f'Unsupported color_format: "{color_format}"')

            heatmap = cv2.resize(heatmap, dsize=size, interpolation=cv2.INTER_CUBIC)

            assert len(heatmap.shape) == 3  # sanity check
            assert (heatmap.shape[1], heatmap.shape[0]) == size  # sanity check
            self.__heatmap_image_by_size_and_color_format[cache_key] = heatmap

        return self.__heatmap_image_by_size_and_color_format[cache_key]

    def _invalidate_cached_computations(self):
        self.__heatmap_image_by_size_and_color_format = {}



"""
(*)~---------------------------------------------------------------------------
Pupil - eye tracking platform
Copyright (C) 2012-2020 Pupil Labs
Distributed under the terms of the GNU
Lesser General Public License (LGPL v3.0).
See LICENSE for license details.
---------------------------------------------------------------------------~(*)
"""
import logging
import typing as T

import numpy as np

from .camera import Camera
from .corner import CornerId
from .heatmap import SurfaceHeatmap
from .image_crop import SurfaceImageCrop
from .location import SurfaceLocation
from .marker import Marker, MarkerId
from .orientation import SurfaceOrientation
from .surface import Surface
from .visual_anchors import SurfaceVisualAnchors

logger = logging.getLogger(__name__)


class SurfaceTracker:
    def __init__(self):
        self.__locations_tracker = _SurfaceTrackerWeakLocationStore()
        self.__argument_validator = _SurfaceTrackerArgumentValidator()

    # ## Creating a surface

    def define_surface(self, name: str, markers: T.List[Marker]) -> T.Optional[Surface]:
        """ """

        return Surface._create_surface_from_markers(name=name, markers=markers)

    # ## Inspecting a surface

    def surface_corner_positions_in_image_space(
        self, surface: Surface, location: SurfaceLocation, corners: T.List[CornerId]
    ) -> T.Mapping[CornerId, T.Tuple[int, int]]:
        """Return the corner positions in image space."""

        # Validate the surface definition and the surface location arguments
        self.__argument_validator.validate_surface_and_location(
            surface=surface, location=location, ignore_location_staleness=False
        )

        if len(corners) == 0:
            return {}

        positions = self.surface_points_in_image_space(
            surface=surface,
            location=location,
            points=[corner.as_tuple() for corner in corners],
        )

        assert len(corners) == len(positions)  # sanity check

        updated_corners = surface.orientation.map_surface_corners_to_rotation(
            dict(zip(corners, positions))
        )

        return updated_corners

    def surface_points_in_image_space(
        self,
        surface: Surface,
        location: SurfaceLocation,
        points: T.List[T.Tuple[float, float]],
    ) -> T.List[T.Tuple[int, int]]:
        """Transform a list of points in surface space into a list of points in image space."""

        # Validate the surface definition and the surface location arguments
        self.__argument_validator.validate_surface_and_location(
            surface=surface, location=location, ignore_location_staleness=False
        )

        if len(points) == 0:
            return []

        return location._map_from_surface_to_image(
            points=np.array(points, dtype=np.float32)
        ).tolist()

    # ## Modifying a surface

    def move_surface_corner_positions_in_image_space(
        self,
        surface: Surface,
        location: SurfaceLocation,
        new_positions: T.Mapping[CornerId, T.Tuple[int, int]],
        ignore_location_staleness: bool = False,
    ):
        """ """

        # Validate the surface definition and the surface location argumentse
        self.__argument_validator.validate_surface_and_location(
            surface=surface,
            location=location,
            ignore_location_staleness=ignore_location_staleness,
        )
        new_positions = surface.orientation.map_surface_corners_to_default(
            new_positions
        )

        if len(new_positions) == 0:
            return

        # Since this action mutates the surface definition, mark previously computed
        # locations as stale
        self.__locations_tracker.mark_locations_as_stale_for_surface(surface=surface)

        ordered_corners = list(new_positions.keys())
        ordered_positions = [new_positions[corner] for corner in ordered_corners]

        ordered_position_in_surface_space_undistorted = (
            location._map_from_image_to_surface(
                points=np.array(ordered_positions, dtype=np.float32)
            ).tolist()
        )

        corner_updates = zip(
            ordered_corners, ordered_position_in_surface_space_undistorted
        )

        for (corner, new_undistorted) in corner_updates:
            # TODO: Provide Surface API for moving multiple corners in one call
            surface._move_corner(
                corner=corner, new_position_in_surface_space_undistorted=new_undistorted
            )

    def add_markers_to_surface(
        self,
        surface: Surface,
        location: SurfaceLocation,
        markers: T.List[Marker],
        ignore_location_staleness: bool = False,
    ):
        """ """

        # Validate the surface definition and the surface location argumentse
        self.__argument_validator.validate_surface_and_location(
            surface=surface,
            location=location,
            ignore_location_staleness=ignore_location_staleness,
        )

        # Ensure marker uniqueness
        marker_uids = {m.uid for m in markers}

        # Ensure only markers that are not part of the definition will be added
        marker_uids = marker_uids.difference(surface.registered_marker_uids)

        # Filter out unused markers
        markers = [m for m in markers if m.uid in marker_uids]

        if len(markers) == 0:
            # If there are no markers to add, return without invalidating the locations
            return

        # Since this action mutates the surface definition, mark previously computed
        # locations as stale
        self.__locations_tracker.mark_locations_as_stale_for_surface(surface=surface)

        for marker in markers:

            marker_undistorted = location._map_marker_from_image_to_surface(
                marker=marker
            )

            surface._add_marker(marker_undistorted=marker_undistorted)

    def remove_markers_from_surface(
        self,
        surface: Surface,
        location: SurfaceLocation,
        marker_uids: T.List[MarkerId],
        ignore_location_staleness: bool = False,
    ):
        """ """

        # Validate the surface definition and the surface location arguments
        self.__argument_validator.validate_surface_and_location(
            surface=surface,
            location=location,
            ignore_location_staleness=ignore_location_staleness,
        )

        # Ensure marker uniqueness
        marker_uids = set(marker_uids)

        # Ensure only markers that are part of the definition will be removed
        marker_uids = marker_uids.intersection(surface.registered_marker_uids)

        if len(marker_uids) == 0:
            # If there are no markers to remove, return without invalidating the
            # locations
            return

        # Since this action mutates the surface definition, mark previously computed
        # locations as stale
        self.__locations_tracker.mark_locations_as_stale_for_surface(surface=surface)

        for marker_uid in marker_uids:
            surface._remove_marker(marker_uid=marker_uid)

    def set_orientation(
        self, surface: Surface, starting_with: CornerId, clockwise: bool = True
    ):
        # Validate the surface definition
        self.__argument_validator.validate_surface(surface=surface)

        orientation = SurfaceOrientation(starting_with, clockwise)
        surface.orientation = orientation

    # ## Locating a surface

    def get_relative_rotation(self, surface: Surface) -> int:
        # Validate the surface definition
        self.__argument_validator.validate_surface(surface=surface)

        return surface.orientation.get_relative_rotation()

    def locate_surface(
        self, surface: Surface, markers: T.List[Marker]
    ) -> T.Optional[SurfaceLocation]:
        """Computes a SurfaceLocation based on a list of visible markers"""

        # Validate the surface definition
        self.__argument_validator.validate_surface(surface=surface)

        location = SurfaceLocation._create_location_from_markers(
            surface=surface, markers=markers
        )

        if location is not None:
            # Track the created location to invalidate it if the surface is mutated
            # later
            self.__locations_tracker.track_location_for_surface(
                surface=surface, location=location
            )

        return location

    def locate_surface_visual_anchors(
        self, surface: Surface, location: SurfaceLocation
    ) -> T.Optional[SurfaceVisualAnchors]:
        """ """

        # Validate the surface definition and the surface location arguments
        self.__argument_validator.validate_surface_and_location(
            surface=surface, location=location, ignore_location_staleness=False
        )

        return SurfaceVisualAnchors._create_from_location(
            location=location, orientation=surface.orientation
        )

    def locate_surface_image_crop(
        self,
        surface: Surface,
        location: SurfaceLocation,
        camera: Camera,
        width: T.Optional[int] = None,
        height: T.Optional[int] = None,
    ) -> SurfaceImageCrop:
        """ """

        # Validate the surface definition and the surface location arguments
        self.__argument_validator.validate_surface_and_location(
            surface=surface, location=location, ignore_location_staleness=False
        )

        # swap width and height if surface location is turned by 90 degree in comparison
        # to internal surface representation
        if surface.orientation.get_relative_rotation() % 180 != 0:
            temp = width
            width = height
            height = temp

        return SurfaceImageCrop._create_image_crop(
            location, camera, width=width, height=height
        )

    def locate_surface_image_crop_with_heatmap(
        self,
        surface: Surface,
        location: SurfaceLocation,
        camera: Camera,
        points: T.List[T.Tuple[int, int]],
        width: T.Optional[int] = None,
        height: T.Optional[int] = None,
    ) -> (SurfaceImageCrop, SurfaceHeatmap):
        """ """

        # Validate the surface definition and the surface location arguments
        self.__argument_validator.validate_surface_and_location(
            surface=surface, location=location, ignore_location_staleness=False
        )

        image_crop = self.locate_surface_image_crop(
            surface=surface,
            location=location,
            camera=camera,
            width=width,
            height=height,
        )

        heatmap = SurfaceHeatmap._create_surface_heatmap(
            points_in_image_space=points, location=location
        )

        return (image_crop, heatmap)


# #### Private Helpers


import weakref


class _SurfaceTrackerWeakLocationStore:
    def __init__(self):
        self.__storage = {}

    def track_location_for_surface(self, surface: Surface, location: SurfaceLocation):
        weak_location = weakref.ref(location)
        surface_locations = self.__storage.get(surface.uid, [])
        surface_locations.append(weak_location)
        self.__storage[surface.uid] = surface_locations

    def mark_locations_as_stale_for_surface(self, surface: Surface):
        # Get all locations that are still being used
        surface_locations = self.__storage.get(surface.uid, [])
        surface_locations = map(lambda l: l(), surface_locations)
        surface_locations = filter(lambda l: l is not None, surface_locations)

        # Invalidate all locations
        for location in surface_locations:
            location._mark_as_stale()

        # Remove locations from tracking
        self.__storage[surface.uid] = []


class _SurfaceTrackerArgumentValidator:
    def __init__(self):
        pass

    @staticmethod
    def validate_surface(surface: Surface):
        """Validate the standalone `surface` argument"""

        if not isinstance(surface, Surface):
            raise ValueError(
                f'Expected an instance of Surface, but got "{surface.__class__}"'
            )

    @staticmethod
    def validate_surface_and_location(
        surface: Surface, location: SurfaceLocation, ignore_location_staleness: bool
    ):
        """Validate the pair of `surface` and `location` arguments"""

        _SurfaceTrackerArgumentValidator.validate_surface(surface=surface)

        if not isinstance(location, SurfaceLocation):
            raise ValueError(
                f'Expected an instance of SurfaceLocation, but got "{location.__class__}"'
            )

        if surface.uid != location.surface_uid:
            raise ValueError(f"SurfaceId missmatch: location doesn't belong to surface")

        if (not ignore_location_staleness) and location.is_stale:
            raise ValueError(
                f"Stale location: the surface definition has changed; location must be recomputed"
            )



"""
(*)~---------------------------------------------------------------------------
Pupil - eye tracking platform
Copyright (C) 2012-2020 Pupil Labs
Distributed under the terms of the GNU
Lesser General Public License (LGPL v3.0).
See LICENSE for license details.
---------------------------------------------------------------------------~(*)
"""
from .corner import CornerId


class SurfaceOrientation:
    def __init__(
        self, surface_start_with: CornerId = CornerId.TOP_LEFT, clockwise: bool = True
    ):
        self.__surface_start_with = surface_start_with
        self.__clockwise = clockwise
        if not clockwise:
            raise NotImplementedError()

    @property
    def surface_start_with(self):
        return self.__surface_start_with

    @property
    def clockwise(self):
        return self.__clockwise

    def map_surface_corners_to_rotation(self, corner_dict):
        mapping = self.__mapping_default_to_rotated()
        return self.__map_surface_corners(corner_dict, mapping)

    def map_surface_corners_to_default(self, corner_dict):
        mapping = self.__mapping_rotated_to_default()
        return self.__map_surface_corners(corner_dict, mapping)

    def __mapping_default_to_rotated(self):
        default_orientation = CornerId.all_corners()
        rotated_orientation = CornerId.all_corners(
            starting_with=self.surface_start_with, clockwise=self.clockwise
        )
        return dict(zip(default_orientation, rotated_orientation))

    def __mapping_rotated_to_default(self):
        default_orientation = CornerId.all_corners()
        rotated_orientation = CornerId.all_corners(
            starting_with=self.surface_start_with, clockwise=self.clockwise
        )
        return dict(zip(rotated_orientation, default_orientation))

    def __map_surface_corners(self, corner_dict, mapping):
        updated_corner_dict = {}
        for corner_name, coords in corner_dict.items():
            updated_corner_dict[mapping[corner_name]] = coords
        return updated_corner_dict

    def get_relative_rotation(self, clockwise=True):
        default_start_with = CornerId.all_corners()[0]
        rotated_orientation = CornerId.all_corners(
            starting_with=self.surface_start_with, clockwise=self.clockwise
        )
        clockwise_rotation = rotated_orientation.index(default_start_with) * 90
        if not clockwise:
            return -clockwise_rotation % 360
        return clockwise_rotation

    def get_visual_anchor_surface_space(self):
        if self.surface_start_with == CornerId.TOP_LEFT and self.clockwise:
            top_indicator_corners = [[0.3, 0.7], [0.7, 0.7], [0.5, 0.9]]
            top_indicator_corners.append(top_indicator_corners[0])
        if self.surface_start_with == CornerId.TOP_RIGHT and self.clockwise:
            top_indicator_corners = [[0.7, 0.7], [0.7, 0.3], [0.9, 0.5]]
            top_indicator_corners.append(top_indicator_corners[0])
        if self.surface_start_with == CornerId.BOTTOM_RIGHT and self.clockwise:
            top_indicator_corners = [[0.7, 0.3], [0.3, 0.3], [0.5, 0.1]]
            top_indicator_corners.append(top_indicator_corners[0])
        if self.surface_start_with == CornerId.BOTTOM_LEFT and self.clockwise:
            top_indicator_corners = [[0.3, 0.3], [0.3, 0.7], [0.1, 0.5]]
            top_indicator_corners.append(top_indicator_corners[0])
        return top_indicator_corners

    def as_dict(self) -> dict:
        return {
            "surface_start_with": self.surface_start_with.name,
            "clockwise": self.clockwise,
        }

    @staticmethod
    def from_dict(value: dict) -> "SurfaceOrientation":
        try:
            return SurfaceOrientation(
                surface_start_with=CornerId.from_name(value["surface_start_with"]),
                clockwise=value["clockwise"],
            )
        except Exception as err:
            raise ValueError(err)



"""
(*)~---------------------------------------------------------------------------
Pupil - eye tracking platform
Copyright (C) 2012-2020 Pupil Labs
Distributed under the terms of the GNU
Lesser General Public License (LGPL v3.0).
See LICENSE for license details.
---------------------------------------------------------------------------~(*)
"""

import cv2
import numpy as np


class Camera:
    def __init__(
        self, camera_matrix: np.ndarray, distortion_coefficients: np.ndarray
    ) -> np.ndarray:
        self.camera_matrix = np.asarray(camera_matrix)
        self.distortion_coefficients = np.asarray(distortion_coefficients)

    def distort_and_project(self, points: np.ndarray):
        points.shape = -1, 1, 2

        points_homogeneous = cv2.convertPointsToHomogeneous(points)
        proj_and_dist_points = np.squeeze(
            cv2.projectPoints(
                points_homogeneous,
                rvec=np.zeros((1, 3)),
                tvec=np.zeros((1, 3)),
                cameraMatrix=self.camera_matrix,
                distCoeffs=self.distortion_coefficients,
            )[0]
        )
        return proj_and_dist_points



"""
(*)~---------------------------------------------------------------------------
Pupil - eye tracking platform
Copyright (C) 2012-2020 Pupil Labs
Distributed under the terms of the GNU
Lesser General Public License (LGPL v3.0).
See LICENSE for license details.
---------------------------------------------------------------------------~(*)
"""
import typing as T

import cv2
import numpy as np

from .camera import Camera
from .corner import CornerId
from .location import SurfaceLocation


# https://www.pyimagesearch.com/2014/08/25/4-point-opencv-getperspective-transform-example/
class SurfaceImageCrop:
    @staticmethod
    def _create_image_crop(
        location: SurfaceLocation,
        camera: Camera,
        width: T.Optional[int],
        height: T.Optional[int],
    ):
        surface_corners_in_surface_space = CornerId.all_corners(
            starting_with=CornerId.TOP_LEFT, clockwise=True
        )
        surface_corners_in_surface_space = [
            c.as_tuple() for c in surface_corners_in_surface_space
        ]
        surface_corners_in_surface_space = np.array(
            surface_corners_in_surface_space, dtype=np.float32
        )

        surface_corners_in_image_space = location._map_from_surface_to_image(
            points=surface_corners_in_surface_space
        )

        surface_corners_in_image_space_distorted = camera.distort_and_project(
            points=surface_corners_in_image_space
        )

        crop_size = SurfaceImageCrop.__calculate_crop_size(
            *surface_corners_in_image_space_distorted, width=width, height=height
        )
        crop_w, crop_h = crop_size

        crop_corners_in_image_space = [
            [0, 0],
            [crop_w - 1, 0],
            [crop_w - 1, crop_h - 1],
            [0, crop_h - 1],
        ]
        crop_corners_in_image_space = np.array(
            crop_corners_in_image_space, dtype=np.float32
        )
        surface_corners_in_image_space_distorted = np.array(
            surface_corners_in_image_space_distorted, dtype=np.float32
        )

        perspective_transform = cv2.getPerspectiveTransform(
            surface_corners_in_image_space_distorted, crop_corners_in_image_space
        )

        return SurfaceImageCrop(
            crop_size_in_image_space=crop_size,
            perspective_transform=perspective_transform,
        )

    def __init__(
        self,
        crop_size_in_image_space: T.Tuple[int, int],
        perspective_transform: np.ndarray,
    ):
        self.__crop_size_in_image_space = crop_size_in_image_space
        self.__perspective_transform = perspective_transform

    @property
    def size_in_image_space(self) -> T.Tuple[int, int]:
        return self.__crop_size_in_image_space

    def apply_to_image(self, image: np.ndarray) -> np.ndarray:
        surface_image = cv2.warpPerspective(
            image, self.__perspective_transform, self.__crop_size_in_image_space
        )
        surface_image = self.__flip_image_vertically(
            surface_image
        )  # Flip image vertically #FIXME: ???
        return surface_image

    @staticmethod
    def __flip_image_vertically(image):
        h = image.shape[0]
        flipped = np.zeros(image.shape, np.uint8)
        for i in range(h):
            flipped[i, :] = image[h - i - 1, :]
        return flipped

    @staticmethod
    def __calculate_crop_size(
        tl: int,
        tr: int,
        br: int,
        bl: int,
        width: T.Optional[int],
        height: T.Optional[int],
    ) -> T.Tuple[int, int]:

        # compute the width of the new image, which will be the
        # maximum distance between bottom-right and bottom-left
        # x-coordiates or the top-right and top-left x-coordinates
        width_a = np.sqrt(((br[0] - bl[0]) ** 2) + ((br[1] - bl[1]) ** 2))
        width_b = np.sqrt(((tr[0] - tl[0]) ** 2) + ((tr[1] - tl[1]) ** 2))
        max_width = max(width_a, width_b)

        # compute the height of the new image, which will be the
        # maximum distance between the top-right and bottom-right
        # y-coordinates or the top-left and bottom-left y-coordinates
        height_a = np.sqrt(((tr[0] - br[0]) ** 2) + ((tr[1] - br[1]) ** 2))
        height_b = np.sqrt(((tl[0] - bl[0]) ** 2) + ((tl[1] - bl[1]) ** 2))
        max_height = max(height_a, height_b)

        # now that we have the dimensions of the new image, construct
        # the set of destination points to obtain a "birds eye view",
        # (i.e. top-down view) of the image, again specifying points
        # in the top-left, top-right, bottom-right, and bottom-left
        # order

        if width is not None and height is not None:
            raise ValueError(f'Expected only "width" OR "height" to be supplied')

        if width is not None:
            ratio = max_width / width
            max_width = width
            max_height /= ratio

        if height is not None:
            ratio = max_height / height
            max_height = height
            max_width /= ratio

        return int(max_width), int(max_height)



"""
(*)~---------------------------------------------------------------------------
Pupil - eye tracking platform
Copyright (C) 2012-2020 Pupil Labs
Distributed under the terms of the GNU
Lesser General Public License (LGPL v3.0).
See LICENSE for license details.
---------------------------------------------------------------------------~(*)
"""
import enum
import typing as T

from . import utils


class CornerId(enum.Enum):
    TOP_LEFT = (0, 0)
    TOP_RIGHT = (1, 0)
    BOTTOM_RIGHT = (1, 1)
    BOTTOM_LEFT = (0, 1)

    @staticmethod
    def all_corners(starting_with: "CornerId" = None, clockwise: bool = True):
        """
        Enumerate all corners.

        Args:
            starting_with: First `CornerId` in the returned list.
            clockwise: Direction of the returned list.

        Returns:
            List of `CornerId`
        """
        starting_with = CornerId.TOP_LEFT if starting_with is None else starting_with

        # Canonical order of the corners
        corners = [
            CornerId.TOP_LEFT,
            CornerId.TOP_RIGHT,
            CornerId.BOTTOM_RIGHT,
            CornerId.BOTTOM_LEFT,
        ]

        if not clockwise:
            corners.reverse()

        n = corners.index(starting_with)
        corners = utils.left_rotation(corners, n)

        return corners

    @staticmethod
    def from_name(name: str):
        return getattr(CornerId, name)

    # ## Serialize

    @staticmethod
    def from_tuple(value: T.Tuple[float, float]):
        return CornerId(value)

    def as_tuple(self) -> T.Tuple[float, float]:
        return tuple(self.value)



"""
(*)~---------------------------------------------------------------------------
Pupil - eye tracking platform
Copyright (C) 2012-2020 Pupil Labs
Distributed under the terms of the GNU
Lesser General Public License (LGPL v3.0).
See LICENSE for license details.
---------------------------------------------------------------------------~(*)
"""
import abc
import typing as T

from .coordinate_space import CoordinateSpace
from .corner import CornerId

MarkerId = T.NewType("MarkerId", str)


class Marker(abc.ABC):

    # ## Abstract members

    @property
    @abc.abstractmethod
    def uid(self) -> MarkerId:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def coordinate_space(self) -> CoordinateSpace:
        raise NotImplementedError()

    def vertices(self) -> T.List[T.Tuple[float, float]]:
        # TODO: Add option to explicitly define the order of the vertices list
        # e.g.: marker.vertices(starting_with=CornerId.BOTTOM_RIGHT, clockwise=False)
        order = CornerId.all_corners()
        return self._vertices_in_order(order=order)

    @abc.abstractmethod
    def _vertices_in_order(
        self, order: T.List[CornerId]
    ) -> T.List[T.Tuple[float, float]]:
        raise NotImplementedError()

    # ## Factory

    @staticmethod
    def from_vertices(
        uid: MarkerId,
        undistorted_image_space_vertices: T.List[T.Tuple[int, int]],
        starting_with: CornerId,
        clockwise: bool,
    ) -> "Marker":
        coordinate_space = CoordinateSpace.IMAGE_UNDISTORTED
        corners = CornerId.all_corners(starting_with=starting_with, clockwise=clockwise)

        expected_len = len(corners)
        actual_len = len(undistorted_image_space_vertices)
        if expected_len != actual_len:
            raise ValueError(
                f'Expected "vertices" to have a lenght of {expected_len}, but got {actual_len}'
            )
        vertices_by_corner_id = dict(zip(corners, undistorted_image_space_vertices))

        return _Marker(
            uid=uid,
            coordinate_space=coordinate_space,
            vertices_by_corner_id=vertices_by_corner_id,
        )

    # ## Serialize

    @staticmethod
    @abc.abstractmethod
    def from_dict(value: dict) -> "Marker":
        return _Marker.from_dict(value)

    @abc.abstractmethod
    def as_dict(self) -> dict:
        raise NotImplementedError()


# #### Concrete implementations


class _Marker(Marker):
    @property
    def uid(self) -> MarkerId:
        return self.__uid

    @property
    def coordinate_space(self) -> CoordinateSpace:
        return self.__coordinate_space

    def _vertices_in_order(
        self, order: T.List[CornerId]
    ) -> T.List[T.Tuple[float, float]]:
        mapping = self.__vertices_by_corner_id
        return [mapping[c] for c in order]

    @staticmethod
    def from_dict(value: dict) -> "Marker":
        try:
            return _Marker(
                uid=value["uid"],
                coordinate_space=value["space"],
                vertices_by_corner_id=value["vertices"],
            )
        except Exception as err:
            raise ValueError(err)

    def as_dict(self) -> dict:
        return {
            "uid": self.__uid,
            "space": self.__coordinate_space,
            "vertices": self.__vertices_by_corner_id,
        }

    def __init__(
        self,
        uid: MarkerId,
        coordinate_space: CoordinateSpace,
        vertices_by_corner_id: T.Mapping[CornerId, T.Tuple[float, float]],
    ):
        self.__uid = uid
        self.__coordinate_space = coordinate_space
        self.__vertices_by_corner_id = vertices_by_corner_id



"""
(*)~---------------------------------------------------------------------------
Pupil - eye tracking platform
Copyright (C) 2012-2020 Pupil Labs
Distributed under the terms of the GNU
Lesser General Public License (LGPL v3.0).
See LICENSE for license details.
---------------------------------------------------------------------------~(*)
"""
import abc
import logging
import typing as T

import cv2
import numpy as np

from .coordinate_space import CoordinateSpace
from .corner import CornerId
from .marker import Marker, _Marker
from .surface import Surface, SurfaceId

logger = logging.getLogger(__name__)


class SurfaceLocation(abc.ABC):

    __versioned_subclasses = {}

    def __init_subclass__(cls):
        storage = SurfaceLocation.__versioned_subclasses
        version = cls.version
        if version is None:
            raise ValueError(
                f'SurfaceLocation subclass {cls.__name__} must overwrite class property "version"'
            )
        if version in storage:
            raise ValueError(
                f"SurfaceLocation subclass {cls.__name__} defines an already registered version {version}"
            )
        storage[version] = cls
        return super().__init_subclass__()

    # ## Abstract members

    version = None  # type: ClassVar[int]

    @property
    @abc.abstractmethod
    def surface_uid(self) -> SurfaceId:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def is_stale(self) -> bool:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def number_of_markers_detected(self) -> int:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def transform_matrix_from_image_to_surface_undistorted(self) -> np.ndarray:
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def transform_matrix_from_surface_to_image_undistorted(self) -> np.ndarray:
        raise NotImplementedError()

    # ## Factory

    @staticmethod
    def _create_location_from_markers(
        surface: Surface, markers: T.List[Marker]
    ) -> "SurfaceLocation":

        if surface is None:
            raise ValueError("Surface is None")

        if not all(
            m.coordinate_space == CoordinateSpace.IMAGE_UNDISTORTED for m in markers
        ):
            raise ValueError("Expected all markers to be in undistorted image space")

        registered_marker_uids = set(surface.registered_marker_uids)

        registered_markers_by_uid_undistorted = (
            surface._registered_markers_by_uid_undistorted
        )

        # Return None for an invalid surface definition
        if len(registered_marker_uids) == 0:
            logger.debug(f"Surface does not have any registered markers that define it")
            return None

        # Get the marker UIDs for the visible markers
        visible_markers_by_uid = {m.uid: m for m in markers}
        if len(visible_markers_by_uid) < len(markers):
            raise ValueError(f"Detected markers must be unique")

        # Only keep the visible marker UIDs for markers that define the surface
        visible_markers_by_uid = {
            uid: m
            for uid, m in visible_markers_by_uid.items()
            if uid in registered_marker_uids
        }

        # If no markers are visible - return
        if len(visible_markers_by_uid) == 0:
            return None

        # Get the set of marker UIDs that are both visible and registered
        matching_marker_uids = set(visible_markers_by_uid.keys())

        marker_vertices_order = CornerId.all_corners()

        visible_vertices_undistorted = np.array(
            [
                visible_markers_by_uid[uid]._vertices_in_order(
                    order=marker_vertices_order
                )
                for uid in matching_marker_uids
            ]
        )
        visible_vertices_undistorted.shape = (-1, 2)

        registered_vertices_undistorted = np.array(
            [
                registered_markers_by_uid_undistorted[uid]._vertices_in_order(
                    order=marker_vertices_order
                )
                for uid in matching_marker_uids
            ]
        )
        registered_vertices_undistorted.shape = (-1, 2)

        (
            transform_matrix_from_image_to_surface_undistorted,
            transform_matrix_from_surface_to_image_undistorted,
        ) = _find_homographies(
            registered_vertices_undistorted, visible_vertices_undistorted
        )

        return _SurfaceLocation_v2(
            surface_uid=surface.uid,
            number_of_markers_detected=len(matching_marker_uids),
            transform_matrix_from_image_to_surface_undistorted=transform_matrix_from_image_to_surface_undistorted,
            transform_matrix_from_surface_to_image_undistorted=transform_matrix_from_surface_to_image_undistorted,
        )

    # ## Mapping

    def _map_from_image_to_surface(
        self, points: np.ndarray, transform_matrix=None
    ) -> np.ndarray:
        return self.__map_points(
            points=points,
            transform_matrix=self.__image_to_surface_transform(
                transform_matrix=transform_matrix
            ),
            custom_transformation=False,
        )

    def _map_from_surface_to_image(
        self, points: np.ndarray, transform_matrix=None
    ) -> np.ndarray:
        return self.__map_points(
            points=points,
            transform_matrix=self.__surface_to_image_transform(
                transform_matrix=transform_matrix
            ),
            custom_transformation=True,
        )

    def _map_marker_from_image_to_surface(
        self, marker: Marker, transform_matrix=None
    ) -> Marker:

        order = CornerId.all_corners()

        vertices_in_image_space = marker._vertices_in_order(order=order)

        vertices_in_image_space_numpy = np.asarray(vertices_in_image_space).reshape(
            (4, 2)
        )
        vertices_in_surface_space_numpy = self._map_from_image_to_surface(
            points=vertices_in_image_space_numpy, transform_matrix=transform_matrix
        )

        vertices_in_surface_space = vertices_in_surface_space_numpy.tolist()

        assert len(order) == len(vertices_in_surface_space), "Sanity check"
        vertices_by_corner_id = dict(zip(order, vertices_in_surface_space))

        return _Marker(
            uid=marker.uid,
            coordinate_space=CoordinateSpace.SURFACE_UNDISTORTED,
            vertices_by_corner_id=vertices_by_corner_id,
        )

    def _map_marker_from_surface_to_image(
        self, marker: Marker, transform_matrix=None
    ) -> Marker:

        order = CornerId.all_corners()

        vertices_in_surface_space = marker._vertices_in_order(order=order)

        vertices_in_surface_space_numpy = np.asarray(vertices_in_surface_space).reshape(
            (4, 2)
        )
        vertices_in_image_space_numpy = self._map_from_surface_to_image(
            points=vertices_in_surface_space_numpy, transform_matrix=transform_matrix
        )

        vertices_in_image_space = vertices_in_image_space_numpy.tolist()

        assert len(order) == len(vertices_in_image_space), "Sanity check"
        vertices_by_corner_id = dict(zip(order, vertices_in_image_space))

        return _Marker(
            uid=marker.uid,
            coordinate_space=CoordinateSpace.IMAGE_UNDISTORTED,
            vertices_by_corner_id=vertices_by_corner_id,
        )

    # ## Serialize

    @abc.abstractmethod
    def as_dict(self) -> dict:
        raise NotImplementedError()

    @staticmethod
    @abc.abstractmethod
    def from_dict(value: dict) -> "SurfaceLocation":
        version = value["version"]
        location_class = SurfaceLocation.__versioned_subclasses.get(version, None)
        if location_class is None:
            raise ValueError(
                f"No concrete implementation for SurfaceLocation version {version}"
            )
        return location_class.from_dict(value)

    # ## Private

    def __image_to_surface_transform(
        self, transform_matrix: T.Optional[np.ndarray]
    ) -> np.ndarray:
        if transform_matrix is not None:
            return transform_matrix
        return self.transform_matrix_from_image_to_surface_undistorted

    def __surface_to_image_transform(
        self, transform_matrix: T.Optional[np.ndarray]
    ) -> np.ndarray:
        if transform_matrix is not None:
            return transform_matrix
        return self.transform_matrix_from_surface_to_image_undistorted

    @staticmethod
    def __map_points(
        points: np.ndarray, transform_matrix: np.ndarray, custom_transformation=False
    ) -> np.ndarray:

        points = np.asarray(points)

        # Validate points

        if len(points.shape) == 1 and points.shape[0] == 2:
            pass
        elif len(points.shape) == 2 and points.shape[1] == 2:
            pass
        else:
            raise ValueError(
                f"Expected points to have shape (2,) or (N, 2), but got {points.shape}"
            )

        # Perspective transform

        shape = points.shape
        points.shape = (-1, 1, 2)
        if custom_transformation:
            points = _perspective_transform(points, transform_matrix)
        else:
            points = cv2.perspectiveTransform(points, transform_matrix)
        points.shape = shape

        return points


# #### Helper functions


def is_clockwise_triangle(points):
    p1, p2, p3 = points
    val = (float(p2[0][1] - p1[0][1]) * (p3[0][0] - p2[0][0])) - (
        float(p2[0][0] - p1[0][0]) * (p3[0][1] - p2[0][1])
    )
    return val > 0


def orientation_quadrangle(points, clockwise=True):
    p0, p1, p2, p3 = points
    return all(
        (
            is_clockwise_triangle((p0, p1, p2)) == clockwise,
            is_clockwise_triangle((p1, p2, p3)) == clockwise,
            is_clockwise_triangle((p2, p3, p0)) == clockwise,
        )
    )


def project_points_pos_z(homogeneous_points, transform_matrix):
    res = []
    for p in homogeneous_points:
        projected_point = np.dot(transform_matrix, p[0])
        if projected_point[2] < 0:
            projected_point[2] = (
                np.linalg.norm([projected_point[0], projected_point[1]]) / 32
            )
        res.append(projected_point)
    return res


def _perspective_transform(points, transform_matrix):
    homogeneous_points = cv2.convertPointsToHomogeneous(points)

    proj_unflipped = project_points_pos_z(homogeneous_points, transform_matrix)
    points_unflipped = cv2.convertPointsFromHomogeneous(np.asarray(proj_unflipped))
    orientation_unflipped = orientation_quadrangle(points_unflipped, clockwise=True)

    proj_flipped = project_points_pos_z(-homogeneous_points, transform_matrix)
    points_flipped = cv2.convertPointsFromHomogeneous(np.asarray(proj_flipped))
    orientation_flipped = orientation_quadrangle(points_flipped, clockwise=True)

    if orientation_unflipped and not orientation_flipped:
        return points_unflipped
    if orientation_flipped and not orientation_unflipped:
        return points_flipped
    if orientation_unflipped and orientation_flipped:

        def min_dist(points):
            return min(np.linalg.norm(p) for p in points)

        if min_dist(points_flipped) < min_dist(points_unflipped):
            return points_flipped
        else:
            return points_unflipped
    return points_unflipped


def _find_homographies(points_A, points_B):
    points_A = points_A.reshape((-1, 1, 2))
    points_B = points_B.reshape((-1, 1, 2))

    B_to_A, mask = cv2.findHomography(points_A, points_B)
    # NOTE: cv2.findHomography(A, B) will not produce the inverse of
    # cv2.findHomography(B, A)! The errors can actually be quite large, resulting in
    # on-screen discrepancies of up to 50 pixels. We try to find the inverse
    # analytically instead with fallbacks.

    try:
        A_to_B = np.linalg.inv(B_to_A)
        return A_to_B, B_to_A
    except np.linalg.LinAlgError:
        logger.debug(
            "Failed to calculate inverse homography with np.inv()! "
            "Trying with np.pinv() instead."
        )

    try:
        A_to_B = np.linalg.pinv(B_to_A)
        return A_to_B, B_to_A
    except np.linalg.LinAlgError:
        logger.warning(
            "Failed to calculate inverse homography with np.pinv()! "
            "Falling back to inaccurate manual computation!"
        )

    A_to_B, mask = cv2.findHomography(points_B, points_A)
    return A_to_B, B_to_A


# #### Concrete implementations


class _SurfaceLocation_v2(SurfaceLocation):

    version = 2  # type: ClassVar[int]

    @property
    def surface_uid(self) -> SurfaceId:
        return self.__surface_uid

    @property
    def is_stale(self) -> bool:
        return self.__is_stale

    @property
    def number_of_markers_detected(self) -> int:
        return self.__number_of_markers_detected

    @property
    def transform_matrix_from_image_to_surface_undistorted(self) -> np.ndarray:
        return self.__transform_matrix_from_image_to_surface_undistorted

    @property
    def transform_matrix_from_surface_to_image_undistorted(self) -> np.ndarray:
        return self.__transform_matrix_from_surface_to_image_undistorted

    def __init__(
        self,
        surface_uid: SurfaceId,
        number_of_markers_detected: int,
        transform_matrix_from_image_to_surface_undistorted: np.ndarray,
        transform_matrix_from_surface_to_image_undistorted: np.ndarray,
    ):
        self.__is_stale = False
        self.__surface_uid = surface_uid
        self.__number_of_markers_detected = number_of_markers_detected
        self.__transform_matrix_from_image_to_surface_undistorted = (
            transform_matrix_from_image_to_surface_undistorted
        )
        self.__transform_matrix_from_surface_to_image_undistorted = (
            transform_matrix_from_surface_to_image_undistorted
        )

    def _mark_as_stale(self):
        self.__is_stale = False

    def as_dict(self) -> dict:
        return {
            "version": self.version,
            "surface_uid": str(self.surface_uid),
            "number_of_markers_detected": self.number_of_markers_detected,
            "transform_matrix_from_image_to_surface_undistorted": self.transform_matrix_from_image_to_surface_undistorted.tolist(),
            "transform_matrix_from_surface_to_image_undistorted": self.transform_matrix_from_surface_to_image_undistorted.tolist(),
        }

    @staticmethod
    def from_dict(value: dict) -> "SurfaceLocation":
        try:
            actual_version = value["version"]
            expected_version = _SurfaceLocation_v2.version
            assert (
                expected_version == actual_version
            ), f"SurfaceLocation version missmatch; expected {expected_version}, but got {actual_version}"
            return _SurfaceLocation_v2(
                surface_uid=SurfaceId(value["surface_uid"]),
                number_of_markers_detected=int(value["number_of_markers_detected"]),
                transform_matrix_from_image_to_surface_undistorted=np.asarray(
                    value["transform_matrix_from_image_to_surface_undistorted"]
                ),
                transform_matrix_from_surface_to_image_undistorted=np.asarray(
                    value["transform_matrix_from_surface_to_image_undistorted"]
                ),
            )
        except Exception as err:
            raise ValueError(err)



"""
(*)~---------------------------------------------------------------------------
Pupil - eye tracking platform
Copyright (C) 2012-2020 Pupil Labs
Distributed under the terms of the GNU
Lesser General Public License (LGPL v3.0).
See LICENSE for license details.
---------------------------------------------------------------------------~(*)
"""


def left_rotation(a: list, k: int):
    """Rotate list to the left
    e.g.: [1, 2, 3, 4] -> [2, 3, 4, 1]
    """
    # if the size of k > len(a), rotate only necessary with
    # # module of the division
    rotations = k % len(a)
    return a[rotations:] + a[:rotations]


def right_rotation(a: list, k: int):
    """Rotate list to the right
    e.g.: [1, 2, 3, 4] -> [4, 1, 2, 3]
    """
    # if the size of k > len(a), rotate only necessary with
    # module of the division
    rotations = k % len(a)
    return a[-rotations:] + a[:-rotations]



"""
(*)~---------------------------------------------------------------------------
Pupil - eye tracking platform
Copyright (C) 2012-2020 Pupil Labs
Distributed under the terms of the GNU
Lesser General Public License (LGPL v3.0).
See LICENSE for license details.
---------------------------------------------------------------------------~(*)
"""
from . import utils
from .camera import Camera
from .coordinate_space import CoordinateSpace
from .corner import CornerId
from .heatmap import SurfaceHeatmap
from .image_crop import SurfaceImageCrop
from .location import SurfaceLocation
from .marker import Marker, MarkerId
from .orientation import SurfaceOrientation
from .surface import Surface, SurfaceId
from .tracker import SurfaceTracker
from .visual_anchors import SurfaceVisualAnchors

__version__ = "0.0.1"



"""
(*)~---------------------------------------------------------------------------
Pupil - eye tracking platform
Copyright (C) 2012-2020 Pupil Labs
Distributed under the terms of the GNU
Lesser General Public License (LGPL v3.0).
See LICENSE for license details.
---------------------------------------------------------------------------~(*)
"""
import enum


class CoordinateSpace(enum.Enum):
    IMAGE_DISTORTED = "image-distorted"
    IMAGE_UNDISTORTED = "image-undistorted"
    SURFACE_DISTORTED = "surface-distorted"
    SURFACE_UNDISTORTED = "surface-undisitorted"



"""
(*)~---------------------------------------------------------------------------
Pupil - eye tracking platform
Copyright (C) 2012-2020 Pupil Labs
Distributed under the terms of the GNU
Lesser General Public License (LGPL v3.0).
See LICENSE for license details.
---------------------------------------------------------------------------~(*)
"""
from setuptools import setup

setup(
    extras_require={
        "dev": ["pre-commit", "tox"],
        "deploy": ["build", "twine", "bump2version"],
        "example": [
            "opencv-python",
            "pupil-apriltags",
            "matplotlib",
            "Pillow",
            "msgpack",
        ],
    }
)



"""
(*)~---------------------------------------------------------------------------
Pupil - eye tracking platform
Copyright (C) 2012-2020 Pupil Labs
Distributed under the terms of the GNU
Lesser General Public License (LGPL v3.0).
See LICENSE for license details.
---------------------------------------------------------------------------~(*)
"""
import typing as T

import numpy as np

from .corner import CornerId
from .location import SurfaceLocation
from .orientation import SurfaceOrientation


class SurfaceVisualAnchors:
    @staticmethod
    def _create_from_location(
        location: SurfaceLocation, orientation: SurfaceOrientation
    ) -> "SurfaceVisualAnchors":

        perimeter_corners = CornerId.all_corners()
        perimeter_corners = [c.as_tuple() for c in perimeter_corners]
        perimeter_corners.append(perimeter_corners[0])

        top_indicator_corners = orientation.get_visual_anchor_surface_space()

        perimeter_points_surface_space = np.array(perimeter_corners, dtype=np.float32)
        perimeter_points_image_space = location._map_from_surface_to_image(
            points=perimeter_points_surface_space
        )

        top_indicator_points_in_surface_space = np.array(
            top_indicator_corners, dtype=np.float32
        )
        top_indicator_points_in_image_space = location._map_from_surface_to_image(
            points=top_indicator_points_in_surface_space
        )

        title_anchor = perimeter_points_image_space.reshape((5, -1))[2]
        title_anchor = title_anchor[0], title_anchor[1] - 75

        edit_surface_anchor = title_anchor[0], title_anchor[1] + 25
        edit_markers_anchor = title_anchor[0], title_anchor[1] + 50

        return SurfaceVisualAnchors(
            top_polyline=top_indicator_points_in_image_space.reshape((4, 2)).tolist(),
            perimeter_polyline=perimeter_points_image_space.reshape((5, 2)).tolist(),
            title_anchor=title_anchor,
            edit_surface_anchor=edit_surface_anchor,
            edit_markers_anchor=edit_markers_anchor,
        )

    def __init__(
        self,
        top_polyline: T.List[T.Tuple[float, float]],
        perimeter_polyline: T.List[T.Tuple[float, float]],
        title_anchor: T.Tuple[float, float],
        edit_surface_anchor: T.Tuple[float, float],
        edit_markers_anchor: T.Tuple[float, float],
    ):
        self.__top_polyline = top_polyline
        self.__perimeter_polyline = perimeter_polyline
        self.__title_anchor = title_anchor
        self.__edit_surface_anchor = edit_surface_anchor
        self.__edit_markers_anchor = edit_markers_anchor

    @property
    def top_polyline(self) -> T.List[T.Tuple[float, float]]:
        return list(self.__top_polyline)

    @property
    def perimeter_polyline(self) -> T.List[T.Tuple[float, float]]:
        return list(self.__perimeter_polyline)

    @property
    def title_anchor(self) -> T.Tuple[float, float]:
        return tuple(self.__title_anchor)

    @property
    def edit_surface_anchor(self) -> T.Tuple[float, float]:
        return tuple(self.__edit_surface_anchor)

    @property
    def edit_markers_anchor(self) -> T.Tuple[float, float]:
        return tuple(self.__edit_markers_anchor)
